import pkg_resources

from . import bloomfilter, field_formats, key_derivation, schema, randomnames, describe
from .schema import Schema

try:
    __version__ = pkg_resources.get_distribution('clkhash').version
except pkg_resources.DistributionNotFound:
    __version__ = "development"

__author__ = "Data61"



from typing import Tuple, Union, Optional, Sequence

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF

"""
We use the block-size of SHA1 and MD5 as the default key size for HMAC
"""
DEFAULT_KEY_SIZE = 64

DEFAULT_NUM_HASHING_METHODS = 2

_HASH_FUNCTIONS = {
    'SHA256': hashes.SHA256,
    'SHA512': hashes.SHA512
}


def hkdf(secret: bytes,
         num_keys: int,
         hash_algo: str = 'SHA256',
         salt: Optional[bytes] = None,
         info: Optional[bytes] = None,
         key_size: int = DEFAULT_KEY_SIZE
         ) -> Tuple[bytes, ...]:
    """
    Executes the HKDF key derivation function as described in rfc5869 to
    derive `num_keys` keys of size `key_size` from the secret.

    :param secret: input keying material
    :param num_keys: the number of keys the kdf should produce
    :param hash_algo: The hash function used by HKDF for the internal
        HMAC calls. The choice of hash function defines the maximum
        length of the output key material. Output bytes <= 255 * hash
        digest size (in bytes).
    :param salt: HKDF is defined to operate with and without random
        salt. This is done to accommodate applications where a salt
        value is not available. We stress, however, that the use of salt
        adds significantly to the strength of HKDF, ensuring
        independence between different uses of the hash function,
        supporting "source-independent" extraction, and strengthening
        the analytical results that back the HKDF design.
        Random salt differs fundamentally from the initial keying
        material in two ways: it is non-secret and can be re-used.
        Ideally, the salt value is a random (or pseudorandom) string
        of the length HashLen.  Yet, even a salt value of less quality
        (shorter in size or with limited entropy) may still make a
        significant contribution to the security of the output keying
        material.
    :param info: While the 'info' value is optional in the definition of
        HKDF, it is often of great importance in applications. Its main
        objective is to bind the derived key material to application-
        and context-specific information. For example, 'info' may
        contain a protocol number, algorithm identifiers, user
        identities, etc.  In particular, it may prevent the derivation
        of the same keying material for different contexts (when the
        same input key material (IKM) is used in such different
        contexts).  It may also accommodate additional inputs to the key
        expansion part, if so desired (e.g., an application may want to
        bind the key material to its length L, thus making L part of the
        'info' field).  There is one technical requirement from 'info':
        it should be independent of the input key material value IKM.
    :param key_size: the size of the produced keys
    :return: Derived keys
    """
    try:
        hash_function = _HASH_FUNCTIONS[hash_algo]
    except KeyError as e:
        msg = "unsupported hash function '{}'".format(hash_algo)
        raise ValueError(msg) from e

    hkdf = HKDF(algorithm=hash_function(),
                length=num_keys * key_size,
                salt=salt,
                info=info,
                backend=default_backend())
    # hkdf.derive returns a block of num_keys * key_size bytes which we
    # divide up into num_keys chunks, each of size key_size
    keybytes = hkdf.derive(secret)
    keys = tuple(keybytes[i * key_size:(i + 1) * key_size] for i in range(num_keys))
    return keys


def generate_key_lists(secret: Union[bytes, str],
                       num_identifier: int,
                       num_hashing_methods: int = DEFAULT_NUM_HASHING_METHODS,
                       key_size: int = DEFAULT_KEY_SIZE,
                       salt: Optional[bytes] = None,
                       info: Optional[bytes] = None,
                       kdf: str = 'HKDF',
                       hash_algo: str = 'SHA256'
                       ) -> Tuple[Tuple[bytes, ...], ...]:
    """
    Generates `num_hashing_methods` derived keys for each identifier for the secret using a key derivation
    function (KDF).

    The only supported key derivation function for now is 'HKDF'.

    The previous secret usage can be reproduced by setting kdf to 'legacy', but it will use the secret twice.
    This is highly discouraged, as this strategy will map the same n-grams in different identifier
    to the same bits in the Bloom filter and thus does not lead to good results.

    :param secret: a secret (either as bytes or string)
    :param num_identifier: the number of identifiers
    :param num_hashing_methods: number of hashing methods used per identifier, each of them requiring a different key
    :param key_size: the size of the derived keys
    :param salt: salt for the KDF as bytes
    :param info: optional context and application specific information as bytes
    :param kdf: the key derivation function algorithm to use
    :param hash_algo: the hashing algorithm to use (ignored if `kdf` is not 'HKDF')
    :return: The derived keys.
             First dimension is of size num_identifier, second dimension is of size num_hashing_methods
             A key is represented as bytes.
    """
    if num_hashing_methods < 1:
        raise ValueError('num_hashing_methods: "{}" is not supported, it'
                         ' should be greater than 0.'.format(num_hashing_methods))
    try:
        if isinstance(secret, bytes):
            secret_bytes = secret
        else:
            secret_bytes = secret.encode('UTF-8')
    except AttributeError:
        raise TypeError("provided 'secret' has to be either of type bytes or strings.")
    if kdf == 'HKDF':
        # we first create the good number of keys, and we then pack them in the expected way.
        key_tuples = hkdf(secret_bytes, num_hashing_methods * num_identifier,
                          hash_algo=hash_algo, salt=salt,
                          info=info, key_size=key_size)
        # regroup such that we get a tuple of keys for each identifier
        split_list = [key_tuples[(i*num_hashing_methods):((i+1)*num_hashing_methods)] for i in range(num_identifier)]
        return tuple(split_list)
    if kdf == 'legacy':
        return tuple(tuple([secret_bytes] * num_hashing_methods) for _ in range(num_identifier))
    raise ValueError('kdf: "{}" is not supported.'.format(kdf))



# -*- coding: utf-8 -*-

""" Schema loading and validation.
"""
import base64
import json
import pkgutil
from typing import Any, Dict, Hashable, Optional, Sequence, Text, TextIO
from copy import deepcopy

import jsonschema

from clkhash.field_formats import FieldSpec, spec_from_json_dict, InvalidSchemaError
from clkhash.key_derivation import DEFAULT_KEY_SIZE as DEFAULT_KDF_KEY_SIZE

MASTER_SCHEMA_FILE_NAMES = {1: 'v1.json',
                            2: 'v2.json',
                            3: 'v3.json'}  # type: Dict[Hashable, Text]


class SchemaError(Exception):
    """ The user-defined schema is invalid.
    """

    def __init__(self,
                 msg: str,
                 errors: Optional[Sequence[InvalidSchemaError]] = None
                 ) -> None:
        self.msg = msg
        self.errors = [] if errors is None else errors
        super(SchemaError, self).__init__(msg)

    def __str__(self) -> str:
        detail = ""
        for i, e in enumerate(self.errors, start=1):
            detail += "Error {} in feature at index {} - {}\n".format(i, e.field_spec_index, str(e))
            detail += "Invalid spec:\n{}\n---\n".format(e.json_field_spec)

        return self.msg + '\n\n' + detail


class MasterSchemaError(Exception):
    """ Master schema missing? Corrupted? Otherwise surprising? This is
        the exception for you!
    """


class Schema:
    """Linkage Schema which describes how to encode plaintext identifiers.

    :ivar fields: the features or field definitions
    :ivar int l: The length of the resulting encoding in bits. This is the length after XOR folding.
    :ivar int xor_folds: The number of XOR folds to perform on the hash.
    :ivar str kdf_type: The key derivation function to use. Currently,
        the only permitted value is 'HKDF'.
    :ivar str kdf_hash: The hash function to use in key derivation. The
        options are 'SHA256' and 'SHA512'.
    :ivar bytes kdf_info: The info for key derivation. See documentation
        of :func:`key_derivation.hkdf` for details.
    :ivar bytes kdf_salt: The salt for key derivation. See documentation
        of :func:`key_derivation.hkdf` for details.
    :ivar int kdf_key_size: The size of the derived keys in bytes.
    """

    def __init__(self,
                 fields: Sequence[FieldSpec],
                 l: int,
                 xor_folds: int = 0,
                 kdf_type: str = 'HKDF',
                 kdf_hash: str = 'SHA256',
                 kdf_info: Optional[bytes] = None,
                 kdf_salt: Optional[bytes] = None,
                 kdf_key_size: int = DEFAULT_KDF_KEY_SIZE
                 ) -> None:
        self.fields = fields
        self.l = l
        self.xor_folds = xor_folds

        self.kdf_type = kdf_type
        self.kdf_type = kdf_type
        self.kdf_hash = kdf_hash
        self.kdf_info = kdf_info
        self.kdf_salt = kdf_salt
        self.kdf_key_size = kdf_key_size

    def __repr__(self):
        return "<Schema (v3): {} fields>".format(len(self.fields))


def _convert_v1_to_v2(schema_dict: Dict[str, Any]) -> Dict[str, Any]:
    """
    Convert v1 schema dict to v2 schema dict.
    :param schema_dict: v1 schema dict
    :return: v2 schema dict
    """
    schema_dict = deepcopy(schema_dict)
    version = schema_dict['version']
    if version != 1:
        raise ValueError('Version {} not 1'.format(version))

    clk_config = schema_dict['clkConfig']
    k = clk_config.pop('k')
    clk_hash = clk_config['hash']

    def convert_feature(f):
        if f.get('ignored', False):
            return f

        hashing = f['hashing']
        weight = hashing.get('weight', 1.0)

        if weight == 0:
            return {
                'identifier': f['identifier'],
                'ignored': True
            }

        x = deepcopy(f)
        hashing = x['hashing']
        if 'weight' in hashing:
            del hashing['weight']

        hashing['strategy'] = {}
        hashing['strategy']['k'] = int(round(weight * k))
        hashing['hash'] = clk_hash
        return x

    result = {
        'version': 2,
        'clkConfig': {
            'l': clk_config['l'],
            'xor_folds': clk_config.get('xor_folds', 0),
            'kdf': clk_config['kdf']
        },
        'features': list(map(convert_feature, schema_dict['features']))
    }
    return result


def _convert_v2_to_v3(schema_dict: Dict[str, Any]) -> Dict[str, Any]:
    """
    Convert v2 schema dict to v3 schema dict.
    :param schema_dict: v2 schema dict
    :return: v3 schema dict
    """
    schema_dict = deepcopy(schema_dict)
    version = schema_dict['version']
    if version != 2:
        raise ValueError('Version {} not 2'.format(version))
    schema_dict['version'] = 3
    for feature in schema_dict['features']:
        if feature.get('ignored', False):
            continue
        strategy = feature['hashing']['strategy']
        if 'k' in strategy:
            strategy['bitsPerToken'] = strategy.pop('k')
        elif 'numBits' in strategy:
            strategy['bitsPerFeature'] = strategy.pop('numBits')
        ngrams = feature['hashing']['ngram']
        feature['hashing']['comparison'] = {'type': 'ngram', 'n': feature['hashing'].pop('ngram'), 'positional': feature['hashing'].pop('positional', False)}
    return schema_dict


def convert_to_latest_version(schema_dict: Dict[str, Any], validate_result: Optional[bool] = False) -> Dict[str, Any]:
    """ Convert the given schema to latest schema version.

     :param schema_dict: A dictionary describing a linkage schema. This dictionary must have a `'version'` key
            containing a master schema version. The rest of the schema dict must conform to the corresponding
            master schema.
     :param validate_result: validate converted schema against schema specification
     :return: schema dict of the latest version
     :raises SchemaError: if schema version is not supported
     """
    version = schema_dict.get('version', "'not specified'")
    if version not in MASTER_SCHEMA_FILE_NAMES.keys():
        msg = ('Schema version {} is not supported. '
               'Consider updating clkhash.').format(version)
        raise SchemaError(msg)
    if schema_dict['version'] == 1:
        schema_dict = _convert_v1_to_v2(schema_dict)
    if schema_dict['version'] == 2:
        schema_dict = _convert_v2_to_v3(schema_dict)
    if validate_result:
        validate_schema_dict(schema_dict)
    return schema_dict


def from_json_dict(dct: Dict[str, Any], validate: bool = True) -> Schema:
    """ Create a Schema of the most recent version according to dct

    if the provided schema dict is of an older version, then it will be automatically converted to the latest.

    :param dct: This dictionary must have a `'features'`
            key specifying the columns of the dataset. It must have
            a `'version'` key containing the master schema version
            that this schema conforms to. It must have a `'hash'`
            key with all the globals.
    :param validate: (default True) Raise an exception if the
            schema does not conform to the master schema.
    :raises SchemaError: An exception containing details about why
            the schema is not valid.
    :return: the Schema
    """
    if validate:
        # This raises iff the schema is invalid.
        validate_schema_dict(dct)
    dct = convert_to_latest_version(dct)
    if validate:
        validate_schema_dict(dct)
    clk_config = dct['clkConfig']
    l = clk_config['l']
    xor_folds = clk_config.get('xor_folds', 0)

    kdf = clk_config['kdf']
    kdf_type = kdf['type']
    kdf_hash = kdf.get('hash', 'SHA256')
    kdf_info_string = kdf.get('info')
    kdf_info = (base64.b64decode(kdf_info_string)
                if kdf_info_string is not None
                else None)
    kdf_salt_string = kdf.get('salt')
    kdf_salt = (base64.b64decode(kdf_salt_string)
                if kdf_salt_string is not None
                else None)
    kdf_key_size = kdf.get('keySize', DEFAULT_KDF_KEY_SIZE)

    # Try to parse each feature config and store any errors encountered
    # for reporting.
    feature_errors = []
    feature_configs = []

    for i, feature_config in enumerate(dct['features']):
        try:
            feature_configs.append(spec_from_json_dict(feature_config))
        except InvalidSchemaError as e:
            e.field_spec_index = i
            e.json_field_spec = feature_config
            feature_errors.append(e)

    if len(feature_errors):
        raise SchemaError("Schema was invalid", feature_errors)

    return Schema(feature_configs, l, xor_folds,
                  kdf_type, kdf_hash, kdf_info, kdf_salt, kdf_key_size)


def from_json_file(schema_file: TextIO, validate: bool = True) -> Schema:
    """ Load a Schema object from a json file.

        :param schema_file: A JSON file containing the schema.
        :param validate: (default True) Raise an exception if the
            schema does not conform to the master schema.
        :raises SchemaError: When the schema is invalid.
        :return: the Schema

    """
    try:
        schema_dict = json.load(schema_file)
    except ValueError as e:  # In Python 3 we can be more specific
        # with json.decoder.JSONDecodeError,
        # but that doesn't exist in Python 2.
        msg = 'The schema is not a valid JSON file.'
        raise SchemaError(msg) from e

    return from_json_dict(schema_dict, validate=validate)


def _get_master_schema(version: Hashable) -> dict:
    """ Loads the master schema of given version

        :param version: The version of the master schema whose path we
            wish to retrieve.
        :raises SchemaError: When the schema version is unknown. This
            usually means that either (a) clkhash is out of date, or (b)
            the schema version listed is incorrect.
        :raises MasterSchemaError: When the master schema is invalid.
        :return: Dict object of the (json) master schema.
    """
    try:
        file_name = MASTER_SCHEMA_FILE_NAMES[version]
    except (TypeError, KeyError) as e:
        msg = ('Schema version {} is not supported. '
               'Consider updating clkhash.').format(version)
        raise SchemaError(msg) from e

    try:
        schema_bytes = pkgutil.get_data('clkhash', 'schemas/{}'.format(file_name))
        if schema_bytes is None:
            msg = ('The master schema could not be loaded. The schema cannot be '
                   'validated. Please file a bug report.')
            raise MasterSchemaError(msg)
    except FileNotFoundError as e:
        msg = ('The master schema could not be found. The schema cannot be '
               'validated. Please file a bug report.')
        raise MasterSchemaError(msg) from e

    try:
        master_schema = json.loads(schema_bytes.decode('utf-8'))
        return master_schema
    except json.decoder.JSONDecodeError as e:
        msg = ('The master schema is not a valid JSON file. The schema cannot '
               'be validated. Please file a bug report.')
        raise MasterSchemaError(msg) from e


def validate_schema_dict(schema: Dict[str, Any]) -> None:
    """ Validate the schema.

        This raises iff either the schema or the master schema are
        invalid. If it's successful, it returns nothing.

        :param schema: The schema to validate, as parsed by `json`.
        :raises SchemaError: When the schema is invalid.
        :raises MasterSchemaError: When the master schema is invalid.
    """
    if not isinstance(schema, dict):
        msg = ('The top level of the schema file is a {}, whereas a dict is '
               'expected.'.format(type(schema).__name__))
        raise SchemaError(msg)

    if 'version' in schema:
        version = schema['version']
    else:
        raise SchemaError('A format version is expected in the schema.')

    master_schema = _get_master_schema(version)

    try:
        jsonschema.validate(schema, master_schema)
    except jsonschema.exceptions.ValidationError as e:
        raise SchemaError('The schema is not valid.\n\n' + str(e)) from e
    except jsonschema.exceptions.SchemaError as e:
        msg = ('The master schema is not valid. The schema cannot be '
               'validated. Please file a bug report.')
        raise MasterSchemaError(msg) from e



# -*- coding: utf-8 -*-


class DescribeError(Exception):
    """ The user provided CLK data is invalid.
    """


def get_encoding_popcounts(clks):
    """

    Often shown as a histogram.

    :param clks: An iterable of CLK encodings.
    :return: An array of integers - the number of bits set for each encoding.
    """

    if len(clks) == 0:
        msg = 'No clks found'
        raise DescribeError(msg)
    try:
        return [clk.count() for clk in clks]
    except Exception as e:
        raise DescribeError("Failed to deserialize encodings") from e



#!/usr/bin/env python3

"""
Serialize bitarray to/from base64 utf-8 encoded string
"""

import base64
from bitarray import bitarray


def serialize_bitarray(ba: bitarray) -> str:
    """Serialize a bitarray (Bloom filter)
    Creates a base64 encoded string representation of the provided bitarray.
    """
    return base64.b64encode(ba.tobytes()).decode('utf8')


def deserialize_bitarray(ser: str) -> bitarray:
    """Deserialize a base 64 encoded string to a bitarray (Bloom filter)
    """
    ba = bitarray()
    ba.frombytes(base64.b64decode(ser.encode(encoding='UTF-8', errors='strict')))
    return ba



# -*- coding: utf-8 -*-

""" Validate data inputs.

    Validates individual entries and the overall format against the
    specified schema.
"""

from typing import cast, Optional, Sequence

from clkhash.field_formats import (FieldSpec, InvalidEntryError)


class EntryError(ValueError):
    """ An entry is invalid.
    """
    row_index = None  # type: Optional[int]
    field_spec = None  # type: Optional[FieldSpec]


class FormatError(ValueError):
    """ The format of the data is invalid.
    """
    row_index = None  # type: Optional[int]


def validate_row_lengths(fields: Sequence[FieldSpec],
                         data: Sequence[Sequence[str]]
                         ) -> None:
    """ Validate the `data` row lengths according to the specification
        in `fields`.

        :param fields: The `FieldSpec` objects forming the
            specification.
        :param data: The rows to check.
        :raises FormatError: When the number of entries in a row does
            not match expectation.
    """
    for i, row in enumerate(data):
        if len(fields) != len(row):
            msg = 'Row {} has {} entries when {} are expected.'.format(
                i, len(row), len(fields))
            raise FormatError(msg)


def validate_entries(fields: Sequence[FieldSpec],
                     data: Sequence[Sequence[str]]
                     ) -> None:
    """ Validate the `data` entries according to the specification in
        `fields`.

        :param fields: The `FieldSpec` objects forming the
            specification.
        :param data: The data to validate.
        :raises EntryError: When an entry is not valid according to its
            :class:`FieldSpec`.
    """
    validators = [f.validate for f in fields]

    for i, row in enumerate(data):
        for entry, v in zip(row, validators):
            try:
                v(entry)
            except InvalidEntryError as e:
                msg = (
                    'Invalid entry in row {row_index}, column '
                    "'{column_name}'. {original_message}"
                ).format(
                    row_index=i,
                    column_name=cast(FieldSpec, e.field_spec).identifier,
                    original_message=e.args[0])
                e_invalid_entry = EntryError(msg)
                e_invalid_entry.field_spec = e.field_spec
                e_invalid_entry.row_index = i
                raise e_invalid_entry from e


def validate_header(fields: Sequence[FieldSpec],
                    column_names: Sequence[str]
                    ) -> None:
    """ Validate the `column_names` according to the specification in
        `fields`.

        :param fields: The `FieldSpec` objects forming the
            specification.
        :param column_names: A sequence of column identifier.
        :raises FormatError: When the number of columns or the column
            identifiers don't match the specification.
    """
    if len(fields) != len(column_names):
        msg = 'Header has {} columns when {} are expected.'.format(
            len(column_names), len(fields))
        raise FormatError(msg)

    for f, column in zip(fields, column_names):
        if f.identifier != column:
            msg = "Column has identifier '{}' when '{}' is expected.".format(
                column, f.identifier)
            raise FormatError(msg)



import os
import tempfile
# noinspection PyProtectedMember
from timeit import default_timer as timer

from clkhash.clk import generate_clk_from_csv
from clkhash.randomnames import NameList


def compute_hash_speed(num: int, quiet: bool = False) -> float:
    """ Hash time.
    """
    namelist = NameList(num)

    os_fd, tmpfile_name = tempfile.mkstemp(text=True)

    schema = NameList.SCHEMA
    header_row = ','.join([f.identifier for f in schema.fields])

    with open(tmpfile_name, 'wt') as f:
        f.write(header_row)
        f.write('\n')
        for person in namelist.names:
            print(','.join([str(field) for field in person]), file=f)

    with open(tmpfile_name, 'rt') as f:
        start = timer()
        generate_clk_from_csv(f, 'secret', schema, progress_bar=not quiet)
        end = timer()

    os.close(os_fd)
    os.remove(tmpfile_name)

    elapsed_time = end - start
    if not quiet:
        print("{:6d} hashes in {:.6f} seconds. {:.2f} KH/s".format(num, elapsed_time, num / (1000 * elapsed_time)))
    return num / elapsed_time


if __name__ == '__main__':
    for n in [100, 1000, 10000, 50000, 100000]:
        compute_hash_speed(n, quiet=n <= 10000)



"""
Generate CLK from data.
"""

import concurrent.futures
import csv
import logging
import time
from typing import (AnyStr, Callable, cast, Iterable, List, Optional,
                    Sequence, TextIO, Tuple, TypeVar, Union)
from bitarray import bitarray
from tqdm import tqdm

from clkhash.bloomfilter import stream_bloom_filters
from clkhash.serialization import serialize_bitarray
from clkhash.key_derivation import generate_key_lists
from clkhash.schema import Schema
from clkhash.stats import OnlineMeanVariance
from clkhash.validate_data import (validate_entries, validate_header,
                                   validate_row_lengths)

log = logging.getLogger('clkhash.clk')

CHUNK_SIZE = 1000


def hash_chunk(chunk_pii_data: Sequence[Sequence[str]],
               keys: Sequence[Sequence[bytes]],
               schema: Schema
               ) -> Tuple[List[bitarray], Sequence[int]]:
    """
    Generate Bloom filters (ie hash) from chunks of PII.
    It also computes and outputs the Hamming weight (or popcount) -- the number of bits
    set to one -- of the generated Bloom filters.

    :param chunk_pii_data: An iterable of indexable records.
    :param keys: A tuple of two lists of keys used in the HMAC. Should have been created by `generate_key_lists`.
    :param Schema schema: Schema specifying the entry formats and
            hashing settings.
    :return: A list of Bloom filters as bitarrays and a list of corresponding popcounts
    """
    clk_data = []
    clk_popcounts = []
    for clk in stream_bloom_filters(chunk_pii_data, keys, schema):
        clk_data.append(clk[0])
        clk_popcounts.append(clk[2])
    return clk_data, clk_popcounts


def generate_clk_from_csv(input_f: TextIO,
                          secret: AnyStr,
                          schema: Schema,
                          validate: bool = True,
                          header: Union[bool, AnyStr] = True,
                          progress_bar: bool = True
                          ) -> List[bitarray]:
    """ Generate Bloom filters from CSV file, then serialise them.

        This function also computes and outputs the Hamming weight
        (a.k.a popcount -- the number of bits set to high) of the
        generated Bloom filters.

        :param input_f: A file-like object of csv data to hash.
        :param secret: A secret.
        :param schema: Schema specifying the record formats and
            hashing settings.
        :param validate: Set to `False` to disable validation of
            data against the schema. Note that this will silence
            warnings whose aim is to keep the hashes consistent between
            data sources; this may affect linkage accuracy.
        :param header: Set to `False` if the CSV file does not have
            a header. Set to `'ignore'` if the CSV file does have a
            header but it should not be checked against the schema.
        :param bool progress_bar: Set to `False` to disable the progress
            bar.
        :return: A list of Bloom filters as bitarrays and a list of
            corresponding popcounts.
    """
    if header not in {False, True, 'ignore'}:
        raise ValueError("header must be False, True or 'ignore' but is {!s}."
                         .format(header))

    log.info("Hashing data")

    # Read from CSV file
    reader = csv.reader(input_f)

    if header:
        column_names = next(reader)
        if header != 'ignore':
            validate_header(schema.fields, column_names)

    start_time = time.time()

    # Read the lines in CSV file and add it to PII
    pii_data = []
    for line in reader:
        pii_data.append(tuple(element.strip() for element in line))

    validate_row_lengths(schema.fields, pii_data)

    if progress_bar:
        stats = OnlineMeanVariance()
        with tqdm(desc="generating CLKs", total=len(pii_data), unit='clk', unit_scale=True,
                  postfix={'mean': stats.mean(), 'std': stats.std()}) as pbar:
            def callback(tics, clk_stats):
                stats.update(clk_stats)
                pbar.set_postfix(mean=stats.mean(), std=stats.std(), refresh=False)
                pbar.update(tics)

            results = generate_clks(pii_data,
                                    schema,
                                    secret,
                                    validate=validate,
                                    callback=callback)
    else:
        results = generate_clks(pii_data,
                                schema,
                                secret,
                                validate=validate)

    log.info("Hashing took {:.2f} seconds".format(time.time() - start_time))
    return results


def generate_clks(pii_data: Sequence[Sequence[str]],
                  schema: Schema,
                  secret: AnyStr,
                  validate: bool = True,
                  callback: Optional[Callable[[int, Sequence[int]], None]] = None
                  ) -> List[bitarray]:

    # Generate two keys for each identifier from the secret, one key per hashing method used when computing
    # the bloom filters.
    # Otherwise it could create more if required using the parameter `num_hashing_methods` in `generate_key_lists`
    key_lists = generate_key_lists(
        secret,
        len(schema.fields),
        key_size=schema.kdf_key_size,
        salt=schema.kdf_salt,
        info=schema.kdf_info,
        kdf=schema.kdf_type,
        hash_algo=schema.kdf_hash)

    if validate:
        validate_entries(schema.fields, pii_data)

    # Chunks PII
    log.info("Hashing {} entities".format(len(pii_data)))
    chunk_size = 200 if len(pii_data) <= 10000 else 1000
    futures = []

    # Compute Bloom filter from the chunks and then serialise it
    with concurrent.futures.ProcessPoolExecutor() as executor:
        for chunk in chunks(pii_data, chunk_size):
            future = executor.submit(
                hash_chunk,
                chunk, key_lists, schema, )
            if callback is not None:
                unpacked_callback = cast(Callable[[int, Sequence[int]], None],
                                         callback)
                future.add_done_callback(
                    lambda f: unpacked_callback(len(f.result()[0]),
                                                f.result()[1]))
            futures.append(future)

        results = []
        for future in futures:
            clks, clk_stats = future.result()
            results.extend(clks)

    return results


T = TypeVar('T')  # Declare generic type variable


def chunks(seq: Sequence[T], chunk_size: int) -> Iterable[Sequence[T]]:
    """ Split seq into chunk_size-sized chunks.

        :param seq: A sequence to chunk.
        :param chunk_size: The size of chunk.
    """
    return (seq[i:i + chunk_size] for i in range(0, len(seq), chunk_size))



"""
Module to produce a dataset of names, genders and dates of birth and manipulate that list

Names and ages are based on Australian and USA census data, but are not correlated.
Additional functions for manipulating the list of names
- producing reordered and subset lists with a specific overlap

ClassList class - generate a list of length n of [id, name, dob, gender] lists

TODO: Generate realistic errors
TODO: Add RESTful api to generate reasonable name data as requested
"""
import bisect
import csv
import json
import math
import pkgutil
import random
from datetime import date, datetime, timedelta
from typing import (Iterable, List, Optional,
                    Sequence, TextIO, Tuple, Union)

from clkhash.field_formats import FieldSpec
from clkhash import schema


def save_csv(data: Iterable[Tuple[Union[str, int], ...]],
             headers: Iterable[str],
             file: TextIO
             ) -> None:
    """
    Output generated data to file as CSV with header.

    :param data: An iterable of tuples containing raw data.
    :param headers: Iterable of feature names
    :param file: A writeable stream in which to write the CSV
    """

    print(','.join(headers), file=file)
    writer = csv.writer(file)
    writer.writerows(data)


def random_date(year: int, age_distribution: Optional['Distribution']) -> datetime:
    """ Generate a random datetime between two datetime objects.

    :param start: datetime of start
    :param end: datetime of end
    :return: random datetime between start and end
    """
    if not age_distribution:
        raise ValueError("Age distribution must be created before creating a random date.")
    try:
        age = int(age_distribution.generate())
    except ValueError:
        raise ValueError("Values in age distribution tables must be integers.")

    start = datetime(year=year - age, month=1, day=1)
    end = datetime(year=year - age + 1, month=1, day=1)

    delta = end - start
    int_delta = (delta.days * 24 * 60 * 60) + delta.seconds
    random_second = random.randrange(int_delta)
    return start + timedelta(seconds=random_second)


class Distribution:
    """Creates a random value generator with a weighted distribution
    """

    def __init__(self, resource_name: str) -> None:
        self.total = 0
        self.indices = []  # type: List[int]
        self.values = []  # type: List[str]
        self.load_csv_data(resource_name)
        self.length = len(self.values)
        if not self.length:
            raise ValueError("Distribution table must have a record.")

    def load_csv_data(self, resource_name: str) -> None:
        """ Loads the first two columns of the specified CSV file from package data.
        The first column represents the value and the second column represents the count in the population.
        """

        data_bytes = pkgutil.get_data('clkhash', '{}'.format(resource_name))
        if not data_bytes:
            raise ValueError("No data resource found with name {}".format(resource_name))
        data = data_bytes.decode('utf8')
        reader = csv.reader(data.splitlines())
        next(reader, None)  # skip the headers
        for row in reader:
            try:
                self.total += int(row[1])
            except ValueError:
                raise ValueError("Distribution resources must only contain integers in the second column.")
            self.indices.append(self.total)
            self.values.append(row[0])

    def generate(self) -> str:
        """ Generates a random value, weighted by the known distribution
        """

        target = random.randint(0, self.total - 1)
        return self.values[bisect.bisect_left(self.indices, target)]


class NameList:
    """ Randomly generated PII records.
    """

    randomname_schema_bytes = pkgutil.get_data('clkhash', 'data/randomnames-schema.json')
    if randomname_schema_bytes is None:
        raise Exception("Couldn't locate package data. Please file a bug report.")
    randomname_schema = json.loads(randomname_schema_bytes.decode())
    SCHEMA = schema.from_json_dict(randomname_schema)

    def __init__(self, n: int) -> None:
        self.load_data()

        self.year = date.today().year - 1

        self.names = [person for person in self.generate_random_person(n)]

        self.all_male_first_names = None  # type: Optional[Distribution]
        self.all_female_first_names = None  # type: Optional[Distribution]
        self.all_last_names = None  # type: Optional[Distribution]
        self.all_ages = None  # type: Optional[Distribution]

    @property
    def schema_types(self) -> Sequence[FieldSpec]:
        return self.SCHEMA.fields

    def generate_random_person(self, n: int) -> Iterable[Tuple[str, str, str, str]]:
        """
        Generator that yields details on a person with plausible name, sex and age.

        :yields: Generated data for one person
            tuple - (id: str, name: str('First Last'), birthdate: str('DD/MM/YYYY'), sex: str('M' | 'F') )
        """
        assert self.all_male_first_names is not None
        assert self.all_female_first_names is not None
        assert self.all_last_names is not None
        for i in range(n):
            sex = 'M' if random.random() > 0.5 else 'F'
            dob = random_date(self.year, self.all_ages).strftime("%Y/%m/%d")
            first_name = self.all_male_first_names.generate() if sex == 'M' else self.all_female_first_names.generate()
            last_name = self.all_last_names.generate()

            yield (
                str(i),
                first_name + ' ' + last_name,
                dob,
                sex
            )

    def load_data(self) -> None:
        """ Loads databases from package data

        Uses data files sourced from
        http://www.quietaffiliate.com/free-first-name-and-last-name-databases-csv-and-sql/
        https://www.census.gov/topics/population/genealogy/data/2010_surnames.html
        https://www.abs.gov.au/AUSSTATS/abs@.nsf/DetailsPage/3101.0Jun%202016
        """

        self.all_male_first_names = Distribution('data/male-first-names.csv')
        self.all_female_first_names = Distribution('data/female-first-names.csv')
        self.all_last_names = Distribution('data/last-names.csv')
        self.all_ages = Distribution('data/ages.csv')

    def generate_subsets(self, sz: int, overlap: float = 0.8, subsets: int = 2) -> Tuple[List, ...]:
        """ Return random subsets with nonempty intersection.

        The random subsets are of specified size. If an element is
        common to two subsets, then it is common to all subsets.
        This overlap is controlled by a parameter.

        :param sz: size of subsets to generate
        :param overlap: size of the intersection, as fraction of the
            subset length
        :param subsets: number of subsets to generate

        :raises ValueError: if there aren't sufficiently many names
            in the list to satisfy the request; more precisely,
            raises if (1 - subsets) * floor(overlap * sz) + subsets * sz > len(self.names).

        :return: tuple of subsets
        """
        overlap_sz = int(math.floor(overlap * sz))
        unique_sz = sz - overlap_sz  # Unique names per subset
        total_unique_sz = unique_sz * subsets  # Uniques in all subsets
        total_sz = overlap_sz + total_unique_sz

        if total_sz > len(self.names):
            msg = 'insufficient names for requested size and overlap'
            raise ValueError(msg)

        sset = random.sample(self.names, total_sz)

        # Overlapping subset, pool of unique names
        sset_overlap, sset_unique = sset[:overlap_sz], sset[overlap_sz:]
        assert len(sset_unique) == subsets * unique_sz

        # Split pool of unique names into `subsets` chunks
        uniques = (sset_unique[p * unique_sz: (p + 1) * unique_sz]
                   for p in range(subsets))

        return tuple(sset_overlap + u for u in uniques)



import math
from typing import Sequence, Union


class OnlineMeanVariance(object):

    def __init__(self) -> None:
        self.t = 0  # type: Union[int, float]
        self.n = 0  # type: int
        self.S = 0  # type: float

    def update(self,
               x: Sequence[Union[int, float]]
               ) -> None:
        """
        updates the statistics with the given list of numbers

        It uses an online algorithm which uses compensated summation to reduce numerical errors.
        See https://angelacorasaniti.wordpress.com/2015/05/06/hw2-mean-and-variance-of-data-stream/ for details.

        :param x: list of numbers
        :return: nothing
        """
        if any(math.isnan(float(i)) or math.isinf(float(i)) for i in x):
            raise ValueError('input contains non-finite numbers like "nan" or "+/- inf"')
        t = sum(x)
        m = float(len(x))
        norm_t = t / m
        S = sum((xi - norm_t) ** 2 for xi in x)
        if self.n == 0:
            self.S = self.S + S
        else:
            self.S = self.S + S + self.n / (m * (m + self.n)) * (m / self.n * self.t - t) ** 2
        self.t = self.t + t
        self.n = self.n + len(x)

    def mean(self) -> float:
        """
        returns the mean

        :return: the mean
        """
        if self.n == 0:
            return 0
        return self.t / float(self.n)

    def variance(self) -> float:
        """
        returns the variance

        :return: the variance
        """
        if self.n <= 1:
            return 0
        return self.S / (self.n - 1.)

    def std(self) -> float:
        """
        returns the standard deviation

        :return: the standard deviation
        """
        return math.sqrt(self.variance())



#!/usr/bin/env python3

"""
Generate a Bloom filter
"""

import hmac
import math
import struct
from hashlib import md5, sha1
from typing import Callable, Iterable, List, Optional, Sequence, Text, Tuple
from bitarray import bitarray

from clkhash.field_formats import FieldHashingProperties
from clkhash.schema import Schema
from clkhash.comparators import AbstractComparison, NonComparison

try:
    from hashlib import blake2b
except ImportError:
    # We are in Python older than 3.6.
    from pyblake2 import blake2b  # type: ignore
    # Ignore because otherwise Mypy raises errors, thinking that
    # blake2b is already defined.


def double_hash_encode_ngrams(ngrams: Iterable[str],
                              keys: Sequence[bytes],
                              ks: Sequence[int],
                              l: int,
                              encoding: str
                              ) -> bitarray:
    """ Computes the double hash encoding of the ngrams with the given keys.

            Using the method from:
            Schnell, R., Bachteler, T., & Reiher, J. (2011).
            A Novel Error-Tolerant Anonymous Linking Code.
            http://grlc.german-microsimulation.de/wp-content/uploads/2017/05/downloadwp-grlc-2011-02.pdf

        :param ngrams: list of n-grams to be encoded
        :param keys: hmac secret keys for md5 and sha1 as bytes
        :param ks: ks[i] is k value to use for ngram[i]
        :param l: length of the output bitarray
        :param encoding: the encoding to use when turning the ngrams to bytes

        :return: bitarray of length l with the bits set which correspond to
                 the encoding of the ngrams
    """
    key_sha1, key_md5 = keys
    bf = bitarray(l)
    bf.setall(False)

    for m, k in zip(ngrams, ks):
        sha1hm = int(
            hmac.new(key_sha1, m.encode(encoding=encoding), sha1).hexdigest(),
            16) % l
        md5hm = int(
            hmac.new(key_md5, m.encode(encoding=encoding), md5).hexdigest(),
            16) % l
        for i in range(k):
            gi = (sha1hm + i * md5hm) % l
            bf[gi] = 1
    return bf


def double_hash_encode_ngrams_non_singular(ngrams: Iterable[str],
                                           keys: Sequence[bytes],
                                           ks: Sequence[int],
                                           l: int,
                                           encoding: str
                                           ) -> bitarray:
    """ computes the double hash encoding of the n-grams with the given keys.

        The original construction of [Schnell2011]_ displays an abnormality for
        certain inputs:

            An n-gram can be encoded into just one bit irrespective of the number
            of k.

        Their construction goes as follows: the :math:`k` different indices
        :math:`g_i` of the Bloom filter for an n-gram
        :math:`x` are defined as:

        .. math:: g_{i}(x) = (h_1(x) + i h_2(x)) \\mod l

        with :math:`0 \\leq i < k` and :math:`l` is the length of the Bloom
        filter. If the value of the hash of :math:`x` of
        the second hash function is a multiple of :math:`l`, then

        .. math:: h_2(x) = 0 \\mod l

        and thus

        .. math:: g_i(x) = h_1(x) \\mod l,

        irrespective of the value :math:`i`. A discussion of this potential flaw
        can be found
        `here <https://github.com/data61/clkhash/issues/33>`_.

        :param ngrams: list of n-grams to be encoded
        :param keys: tuple with (key_sha1, key_md5).
               That is, (hmac secret keys for sha1 as bytes, hmac secret keys for
               md5 as bytes)
        :param ks: ks[i] is k value to use for ngram[i]
        :param l: length of the output bitarray
        :param encoding: the encoding to use when turning the ngrams to bytes

        :return: bitarray of length l with the bits set which correspond to the
                 encoding of the ngrams
    """
    key_sha1, key_md5 = keys
    bf = bitarray(l)
    bf.setall(False)
    for m, k in zip(ngrams, ks):
        m_bytes = m.encode(encoding=encoding)

        sha1hm_bytes = hmac.new(key_sha1, m_bytes, sha1).digest()
        md5hm_bytes = hmac.new(key_md5, m_bytes, md5).digest()

        sha1hm = int.from_bytes(sha1hm_bytes, 'big') % l
        md5hm = int.from_bytes(md5hm_bytes, 'big') % l

        i = 0
        while md5hm == 0:
            md5hm_bytes = hmac.new(
                key_md5, m_bytes + chr(i).encode(), md5).digest()
            md5hm = int.from_bytes(md5hm_bytes, 'big') % l
            i += 1

        for i in range(k):
            gi = (sha1hm + i * md5hm) % l
            bf[gi] = True
    return bf


def blake_encode_ngrams(ngrams: Iterable[str],
                        keys: Sequence[bytes],
                        ks: Sequence[int],
                        l: int,
                        encoding: str
                        ) -> bitarray:
    """ Computes the encoding of the ngrams using the BLAKE2 hash function.

        We deliberately do not use the double hashing scheme as proposed in [
        Schnell2011]_, because this
        would introduce an exploitable structure into the Bloom filter. For more
        details on the
        weakness, see [Kroll2015]_.

        In short, the double hashing scheme only allows for :math:`l^2`
        different encodings for any possible n-gram,
        whereas the use of :math:`k` different independent hash functions gives
        you :math:`\\sum_{j=1}^{k}{\\binom{l}{j}}`
        combinations.


        **Our construction**

        It is advantageous to construct Bloom filters using a family of hash
        functions with the property of
        `k-independence <https://en.wikipedia.org/wiki/K-independent_hashing>`_
        to compute the indices for an entry.
        This approach minimises the change of collisions.

        An informal definition of *k-independence* of a family of hash functions
        is, that if selecting a function at random
        from the family, it guarantees that the hash codes of any designated k
        keys are independent random variables.

        Our construction utilises the fact that the output bits of a
        cryptographic hash function are uniformly distributed,
        independent, binary random variables (well, at least as close to as
        possible. See [Kaminsky2011]_ for an analysis).
        Thus, slicing the output of a cryptographic hash function into k
        different slices gives you k independent random
        variables.

        We chose Blake2 as the cryptographic hash function mainly for two reasons:

        * it is fast.
        * in keyed hashing mode, Blake2 provides MACs with just one hash
          function call instead of the two calls in the HMAC construction used
          in the double hashing scheme.


        .. warning::
           Please be aware that, although this construction makes the attack of
           [Kroll2015]_ infeasible, it is most likely
           not enough to ensure security. Or in their own words:

             | However, we think that using independent hash functions alone
               will not be sufficient to ensure security,
               since in this case other approaches (maybe related to or at least
               inspired through work from the
               area of Frequent Itemset Mining) are promising to detect at least
               the most frequent atoms automatically.

        :param ngrams: list of n-grams to be encoded
        :param keys: secret key for blake2 as bytes
        :param ks: ks[i] is k value to use for ngram[i]
        :param l: length of the output bitarray (has to be a power of 2)
        :param encoding: the encoding to use when turning the ngrams to bytes

        :return: bitarray of length l with the bits set which correspond to the
                 encoding of the ngrams
    """
    key = keys[0]  # We only need the first key

    log_l = int(math.log(l, 2))
    if not 2 ** log_l == l:
        raise ValueError(
            'parameter "l" has to be a power of two for the BLAKE2 encoding, '
            'but was: {}'.format(
                l))
    bf = bitarray(l)
    bf.setall(False)

    for m, k in zip(ngrams, ks):
        random_shorts = []  # type: List[int]
        num_macs = (k + 31) // 32
        for i in range(num_macs):
            hash_bytes = blake2b(m.encode(encoding=encoding), key=key,
                                 salt=str(i).encode()).digest()
            random_shorts.extend(struct.unpack('32H',
                                               hash_bytes))  # interpret
            # hash bytes as 32 unsigned shorts.
        for i in range(k):
            idx = random_shorts[i] % l
            bf[idx] = 1
    return bf


def hashing_function_from_properties(
        fhp: FieldHashingProperties
        ) -> Callable[[Iterable[str], Sequence[bytes], Sequence[int], int, str], bitarray]:
    """ Get the hashing function for this field
        :param fhp: hashing properties for this field
        :return: the hashing function
    """
    if fhp.hash_type == 'doubleHash':
        if fhp.prevent_singularity:
            return double_hash_encode_ngrams_non_singular
        else:
            return double_hash_encode_ngrams
    elif fhp.hash_type == 'blakeHash':
        return blake_encode_ngrams
    else:
        msg = "Unsupported hash type '{}'".format(fhp.hash_type)
        raise ValueError(msg)


def fold_xor(bloomfilter: bitarray,
             folds: int
             ) -> bitarray:
    """ Performs XOR folding on a Bloom filter.

        If the length of the original Bloom filter is n and we perform
        r folds, then the length of the resulting filter is n / 2 ** r.

        :param bloomfilter: Bloom filter to fold
        :param folds: number of folds

        :return: folded bloom filter
    """

    if len(bloomfilter) % 2 ** folds != 0:
        msg = ('The length of the bloom filter is {length}. It is not '
               'divisible by 2 ** {folds}, so it cannot be folded {folds} '
               'times.'
               .format(length=len(bloomfilter), folds=folds))
        raise ValueError(msg)

    for _ in range(folds):
        bf1 = bloomfilter[:len(bloomfilter) // 2]
        bf2 = bloomfilter[len(bloomfilter) // 2:]

        bloomfilter = bf1 ^ bf2

    return bloomfilter


def crypto_bloom_filter(record: Sequence[Text],
                        comparators: List[AbstractComparison],
                        schema: Schema,
                        keys: Sequence[Sequence[bytes]],
                        ) -> Tuple[bitarray, Text, int]:
    """ Computes the composite Bloom filter encoding of a record.

    Using the method from
    http://www.record-linkage.de/-download=wp-grlc-2011-02.pdf

    :param record: plaintext record tuple. E.g. (index, name, dob, gender)
    :param comparators: A list of comparators. They provide a 'tokenize' function to turn string into
        appropriate tokens.
    :param schema: Schema
    :param keys: Keys for the hash functions as a tuple of lists of bytes.

    :return: 3-tuple:

        - bloom filter for record as a bitarray
        - first element of record (usually an index)
        - number of bits set in the bloomfilter
    """
    hash_l = schema.l * 2 ** schema.xor_folds

    bloomfilter = bitarray(hash_l)
    bloomfilter.setall(False)

    for (entry, comparator, field, key) \
            in zip(record, comparators, schema.fields, keys):
        fhp = field.hashing_properties
        if fhp:
            ngrams = list(comparator.tokenize(field.format_value(entry)))
            hash_function = hashing_function_from_properties(fhp)
            if ngrams:
                bloomfilter |= hash_function(ngrams, key,
                                             fhp.strategy.bits_per_token(len(ngrams)),
                                             hash_l, fhp.encoding)

    bloomfilter = fold_xor(bloomfilter, schema.xor_folds)
    return bloomfilter, record[0], bloomfilter.count()


def stream_bloom_filters(dataset: Iterable[Sequence[Text]],
                         keys: Sequence[Sequence[bytes]],
                         schema: Schema
                         ) -> Iterable[Tuple[bitarray, Text, int]]:
    """ Compute composite Bloom filters (CLKs) for every record in an
        iterable dataset.

        :param dataset: An iterable of indexable records.
        :param schema: An instantiated Schema instance
        :param keys: A tuple of two lists of secret keys used in the HMAC.
        :return: Generator yielding bloom filters as 3-tuples
    """
    comparators = [field.hashing_properties.comparator if field.hashing_properties is not None else NonComparison()
                  for field in schema.fields]
    return (crypto_bloom_filter(s, comparators, schema, keys)
            for s in dataset)



# -*- coding: utf-8 -*-

""" Classes that specify the requirements for each column in a dataset.
    They take care of validation, and produce the settings required to
    perform the hashing.
"""

import abc
import re
from datetime import datetime
from typing import Any, Dict, Iterable, Optional, Text, cast, List, Union, SupportsInt

from clkhash import comparators
from clkhash.comparators import AbstractComparison



class InvalidEntryError(ValueError):
    """ An entry in the data file does not conform to the schema.
    """
    field_spec = None  # type: Optional['FieldSpec']


class InvalidSchemaError(ValueError):
    """Raised if the schema of a field specification is invalid.

    For example, a regular expression included in the schema is not
    syntactically correct.
    """
    json_field_spec = None  # type: Optional[dict]
    field_spec_index = None  # type: Optional[int]


class MissingValueSpec(object):
    """ Stores the information about how to find and treat missing
        values.

        :ivar str sentinel: sentinel is the string that identifies a
            missing value e.g.: 'N/A', ''.
            The sentinel will not be validated against the
            feature format definition
        :ivar str replace_with: defines the string which replaces the
            sentinel whenever present, can be 'None', then sentinel will
            not be replaced.

    """

    def __init__(self,
                 sentinel: str,
                 replace_with: Optional[str] = None
                 ) -> None:
        self.sentinel = sentinel
        self.replace_with = (replace_with if replace_with is not None
                             else sentinel)

    @classmethod
    def from_json_dict(cls, json_dict: Dict[str, Any]) -> 'MissingValueSpec':
        return cls(
            sentinel=json_dict['sentinel'],
            replace_with=cast(Optional[str], json_dict.get('replaceWith'))
        )


class StrategySpec(object, metaclass=abc.ABCMeta):
    """ Stores the information about the insertion strategy.

    A strategy has to implement the 'bits_per_token' function, which defines how often each token gets inserted into
    the Bloom filter.
    """

    @abc.abstractmethod
    def bits_per_token(self, num_tokens: int) -> List[int]:
        """ Return a list of integers, one for each of the `num_tokens` tokens, defining how often that token gets
        inserted into the Bloom filter.

        :param int num_tokens: number of tokens in the feature's value
        :return: [ k, ... ] with k's >= 0
        """
        pass

    @classmethod
    def from_json_dict(cls, json_dict: Dict[str, Union[str, SupportsInt]]) -> 'StrategySpec':
        if 'bitsPerToken' in json_dict:
            return BitsPerTokenStrategy(int(json_dict['bitsPerToken']))
        elif 'bitsPerFeature' in json_dict:
            return BitsPerFeatureStrategy(int(json_dict['bitsPerFeature']))
        else:
            raise InvalidSchemaError('unknown strategy')


class BitsPerTokenStrategy(StrategySpec):
    """ Insert every token the same number of times.

    This is the strategy from the original Schnell paper. The provided value `bits_per_token` (the 'k' value in the
    paper) defines the number of hash functions that are used to insert each token into the Bloom filter.

    One important property of this strategy is that the total number of inserted bits for a feature relates to the
    length of its value. This can have privacy implications, as the number of bits set in a Bloom filter correlate to
    the number of tokens of the PII.

    :ivar int bits_per_token: how often each token should be inserted into the filter
    """

    def __init__(self,
                 bits_per_token: int
                 ) -> None:
        self._bits_per_token = bits_per_token

    def bits_per_token(self, num_tokens: int) -> List[int]:
        return [self._bits_per_token] * num_tokens


class BitsPerFeatureStrategy(StrategySpec):
    """ Have a fixed number of filter insertions for a feature, irrespective of the actual number of tokens.

    This strategy allows to reason about the importance of a feature, irrespective of the lengths of the feature values.
    For example, in the BitsPerTokenStrategy the name 'Bob' affects only have the number of bits in the Bloom filter
    than 'Robert'. With this BitsPerFeatureStrategy, both names set the same number of bits in the filter, thus allowing
    to adjust importance on a per feature basis.

    :ivar int bits_per_feature: total number of insertions for this feature, will be spread across all tokens.
    """

    def __init__(self,
                 bits_per_feature: int
                 ) -> None:
        self._bits_per_feature = bits_per_feature

    def bits_per_token(self, num_tokens: int) -> List[int]:
        k = int(self._bits_per_feature / num_tokens)
        residue = self._bits_per_feature % num_tokens
        return ([k + 1] * residue) + ([k] * (num_tokens - residue))


class FieldHashingProperties(object):
    """
    Stores the settings used to hash a field.

    This includes the encoding and tokenisation parameters.

    :ivar AbstractComparison comparator: provides a tokenizer for desired comparison strategy
    :ivar str encoding: The encoding to use when converting the
        string to bytes. Refer to
        `Python's documentation <https://docs.python.org/3/library/codecs.html#standard-encodings>`_
        for possible values.
    :ivar str hash_type: hash function to use for hashing
    :ivar bool prevent_singularity: the 'doubleHash' function has a singularity problem
    :ivar int num_bits: dynamic k = num_bits / number of n-grams
    :ivar int k: max number of bits per n-gram
    :ivar MissingValueSpec missing_value: specifies how to handle missing values
    """
    _DEFAULT_ENCODING = 'utf-8'
    _DEFAULT_POSITIONAL = False

    def __init__(self,
                 comparator: AbstractComparison,
                 strategy: StrategySpec,
                 encoding: str = _DEFAULT_ENCODING,
                 hash_type: str = 'blakeHash',
                 prevent_singularity: Optional[bool] = None,
                 missing_value: Optional[MissingValueSpec] = None
                 ) -> None:
        """ Make a :class:`FieldHashingProperties` object, setting it
            attributes to values specified in keyword arguments.
        """
        if comparator is None:
            raise ValueError('no comparator specified')

        try:
            ''.encode(encoding)
        except LookupError as e:
            msg = '{} is not a valid Python encoding.'
            raise ValueError(msg.format(encoding)) from e

        if prevent_singularity is not None and hash_type != 'doubleHash':
            raise ValueError("Prevent_singularity must only be specified"
                             " with hash_type doubleHash.")

        if strategy is None:
            raise ValueError('no strategy specified')

        self.comparator = comparator
        self.encoding = encoding
        self.hash_type = hash_type
        self.prevent_singularity = prevent_singularity
        self.strategy = strategy
        self.missing_value = missing_value

    def replace_missing_value(self, str_in: Text) -> Text:
        """ returns 'str_in' if it is not equals to the 'sentinel' as
        defined in the missingValue section of
        the schema. Else it will return the 'replaceWith' value.

        :param str str_in: input string
        :return: str_in or the missingValue replacement value
        """
        if self.missing_value is None:
            return str_in
        elif self.missing_value.sentinel == str_in:
            return self.missing_value.replace_with
        else:
            return str_in


def fhp_from_json_dict(
        json_dict: Dict[str, Any]
) -> FieldHashingProperties:
    """
    Make a :class:`FieldHashingProperties` object from a dictionary.

    :param dict json_dict:
        Conforming to the `hashingConfig` definition
        in the `v2` linkage schema.
    :return: A :class:`FieldHashingProperties` instance.
    """
    h = json_dict.get('hash', {'type': 'blakeHash'})

    if json_dict['comparison'].get('type', '') == 'ngram':  # setting default
        json_dict['comparison'].setdefault('positional', FieldHashingProperties._DEFAULT_POSITIONAL)
    comparator = comparators.get_comparator(json_dict['comparison'])

    return FieldHashingProperties(
        comparator=comparator,
        hash_type=h['type'],
        prevent_singularity=h.get('prevent_singularity'),
        strategy=StrategySpec.from_json_dict(json_dict['strategy']),
        missing_value=MissingValueSpec.from_json_dict(
            json_dict[
                'missingValue']) if 'missingValue' in json_dict else None
    )


class FieldSpec(object, metaclass=abc.ABCMeta):
    """ Abstract base class representing the specification of a column
        in the dataset. Subclasses validate entries, and modify the
        `hashing_properties` ivar to customise hashing procedures.

        :ivar str identifier: The name of the field.
        :ivar str description: Description of the field format.
        :ivar FieldHashingProperties hashing_properties: The properties
            for hashing. None if field ignored.
    """

    def __init__(self,
                 identifier: str,
                 hashing_properties: Optional[FieldHashingProperties],
                 description: Optional[str] = None
                 ) -> None:
        """ Make a FieldSpec object, setting it attributes to values
            specified in keyword arguments.
        """
        self.identifier = identifier
        self.hashing_properties = hashing_properties
        self.description = description

    @classmethod
    def from_json_dict(cls,
                       field_dict: Dict[str, Any]
                       ) -> 'FieldSpec':
        """ Initialise a :class:`FieldSpec` object from a dictionary of
            properties.

            :param dict field_dict: The properties dictionary to use. Must
                contain a `'hashing'` key that meets the requirements of
                :class:`FieldHashingProperties`.
            :raises InvalidSchemaError: When the `properties`
                dictionary contains invalid values. Exactly what that
                means is decided by the subclasses.
        """
        identifier = field_dict['identifier']
        description = field_dict['format'].get('description')
        hashing_properties = fhp_from_json_dict(field_dict['hashing']) if 'hashing' in field_dict else None

        result = cls.__new__(cls)
        result.identifier = identifier
        result.hashing_properties = hashing_properties
        result.description = description

        return result

    @abc.abstractmethod
    def validate(self, str_in: Text) -> None:
        """ Validates an entry in the field.

            Raises :class:`InvalidEntryError` iff the entry is invalid.

            Subclasses must override this method with their own
            validation. They should call the parent's `validate` method
            via `super`.

            :param str str_in: String to validate.
            :raises InvalidEntryError: When entry is invalid.
        """
        if self.hashing_properties:  # else its Ignore
            try:
                str_in.encode(encoding=self.hashing_properties.encoding)
            except UnicodeEncodeError as err:
                msg = ("Expected entry that can be encoded in {}. Read '{}'."
                       .format(self.hashing_properties.encoding, str_in))
                e_new = InvalidEntryError(msg)
                e_new.field_spec = self
                raise e_new from err

    def is_missing_value(self, str_in: Text) -> bool:
        """ tests if 'str_in' is the sentinel value for this field

        :param str str_in: String to test if it stands for missing value
        :return: True if a missing value is defined for this field and
            str_in matches this value

        """
        return (self.hashing_properties is not None and
                self.hashing_properties.missing_value is not None and
                self.hashing_properties.missing_value.sentinel == str_in)

    def format_value(self, str_in: Text) -> Text:
        """ formats the value 'str_in' for hashing according to this field's
        spec.

        There are several reasons why this might be necessary:

        1. This field contains missing values which have to be replaced
           by some other string
        2. There are several different ways to describe a specific value
           for this field, e.g.: all of '+65', ' 65',
           '65' are valid representations of the integer 65.
        3. Entries of this field might contain elements with no entropy,
           e.g. dates might be formatted as
           yyyy-mm-dd, thus all dates will have '-' at the same place.
           These artifacts have no value for entity
           resolution and should be removed.

        :param str str_in: the string to format
        :return: a string representation of 'str_in' which is ready to be
            hashed

        """
        if self.hashing_properties and self.is_missing_value(str_in):
            return self.hashing_properties.replace_missing_value(str_in)
        else:
            return self._format_regular_value(str_in)

    def _format_regular_value(self, str_in: Text) -> Text:
        """ overwrite this if you want to modify 'str_in' before hashing.

        :param str str_in:
        :return: a string representation of 'str_in' which is ready to
            be hashed

        """
        return str_in


class StringSpec(FieldSpec):
    """ Represents a field that holds strings.

        One way to specify the format of the entries is to provide a
        regular expression that they must conform to. Another is to
        provide zero or more of: minimum length, maximum length, casing
        (lower, upper, mixed).

        Each string field also specifies an encoding used when turning
        characters into bytes. This is stored in `hashing_properties`
        since it is needed for hashing.

        :ivar str encoding: The encoding to use when converting the
            string to bytes. Refer to
            `Python's documentation
            <https://docs.python.org/3/library/codecs.html#standard
            -encodings>`_
            for possible values.
        :ivar regex: Compiled regular expression that entries must
            conform to. Present only if the specification is regex-
            based.
        :ivar str case: The casing of the entries. One of `'lower'`,
            `'upper'`, or `'mixed'`. Default is `'mixed'`. Present only
            if the specification is not regex-based.
        :ivar int min_length: The minimum length of the string. `None`
            if there is no minimum length. Present only if the
            specification is not regex-based.
        :ivar int max_length: The maximum length of the string. `None`
            if there is no maximum length. Present only if the
            specification is not regex-based.
    """
    _DEFAULT_CASE = 'mixed'
    _DEFAULT_MIN_LENGTH = 0
    _PERMITTED_CASE_STYLES = {'lower', 'upper', 'mixed'}

    def __init__(self,
                 identifier: str,
                 hashing_properties: FieldHashingProperties,
                 description: Optional[str] = None,
                 regex: Optional[str] = None,
                 case: str = _DEFAULT_CASE,
                 min_length: int = _DEFAULT_MIN_LENGTH,
                 max_length: Optional[int] = None
                 ) -> None:
        """ Make a StringSpec object, setting it attributes to values
            specified in keyword arguments.
        """
        # noinspection PyCompatibility,PyArgumentList
        super().__init__(identifier=identifier,
                         description=description,
                         hashing_properties=hashing_properties)

        regex_based = regex is not None

        if regex_based and (case != self._DEFAULT_CASE
                            or min_length != self._DEFAULT_MIN_LENGTH
                            or max_length is not None):
            msg = ('regex cannot be passed along with case, min_length, or'
                   ' max_length.')
            raise ValueError(msg)

        if case not in self._PERMITTED_CASE_STYLES:
            msg = ("the case is {}, but should be 'lower', 'upper', or"
                   "'mixed'")
            raise ValueError(msg.format(case))

        if min_length < 0:
            msg = 'min_length must be non-negative, but is {}'
            raise ValueError(msg.format(min_length))

        # type checker thinks max_length is of type None
        # noinspection PyTypeChecker
        if max_length is not None and max_length <= 0:
            msg = 'max_length must be positive, but is {}'
            raise ValueError(msg.format(max_length))

        if regex_based:
            regex_str = cast(str, regex)
            try:
                self.regex = re.compile(regex_str)
            except (SyntaxError, re.error) as e:
                msg = "invalid regular expression '{}.'".format(regex_str)
                e_new = InvalidEntryError(msg)
                e_new.field_spec = self
                raise e_new from e
        else:
            self.case = case
            self.min_length = min_length
            self.max_length = max_length

        self.regex_based = regex_based

    @classmethod
    def from_json_dict(cls,
                       json_dict: Dict[str, Any]
                       ) -> 'StringSpec':
        """ Make a StringSpec object from a dictionary containing its
            properties.

            :param dict json_dict: This dictionary must contain an
                `'encoding'` key associated with a Python-conformant
                encoding. It must also contain a `'hashing'` key, whose
                contents are passed to :class:`FieldHashingProperties`.
                Permitted keys also include `'pattern'`, `'case'`,
                `'minLength'`, and `'maxLength'`.
            :raises InvalidSchemaError: When a regular expression is
                provided but is not a valid pattern.
        """
        # noinspection PyCompatibility
        result = cast(StringSpec,  # Go away, Mypy.
                      super().from_json_dict(json_dict))

        format_ = json_dict['format']
        if 'encoding' in format_ and result.hashing_properties:
            result.hashing_properties.encoding = format_['encoding']

        if 'pattern' in format_:
            pattern = format_['pattern']
            try:
                result.regex = re.compile(pattern)
            except (SyntaxError, re.error) as e:
                msg = "Invalid regular expression '{}.'".format(pattern)
                e_new = InvalidSchemaError(msg)
                e_new.json_field_spec = json_dict
                raise e_new from e
            result.regex_based = True

        else:
            result.case = format_.get('case', StringSpec._DEFAULT_CASE)
            result.min_length = format_.get('minLength')
            result.max_length = format_.get('maxLength')
            result.regex_based = False

        return result

    def validate(self, str_in: Text) -> None:
        """ Validates an entry in the field.

            Raises `InvalidEntryError` iff the entry is invalid.

            An entry is invalid iff (1) a pattern is part of the
            specification of the field and the string does not match
            it; (2) the string does not match the provided casing,
            minimum length, or maximum length; or (3) the specified
            encoding cannot represent the string.

            :param str str_in: String to validate.
            :raises InvalidEntryError: When entry is invalid.
            :raises ValueError: When self.case is not one of the
                permitted values (`'lower'`, `'upper'`, or `'mixed'`).
        """
        if self.is_missing_value(str_in):
            return
        # noinspection PyCompatibility
        super().validate(str_in)  # Validate encoding.

        if self.regex_based:
            match = self.regex.fullmatch(str_in)
            if match is None:
                e = InvalidEntryError(
                    'Expected entry that conforms to regular expression '
                    "'{}'. Read '{}'.".format(self.regex.pattern, str_in))
                e.field_spec = self
                raise e

        else:
            str_len = len(str_in)
            if self.min_length is not None and str_len < self.min_length:
                e = InvalidEntryError(
                    "Expected string length of at least {}. Read string '{}' "
                    'of length {}.'.format(self.min_length, str_in, str_len))
                e.field_spec = self
                raise e

            if self.max_length is not None and str_len > self.max_length:
                e = InvalidEntryError(
                    "Expected string length of at most {}. Read string '{}' "
                    'of length {}.'.format(self.max_length, str_in, str_len))
                e.field_spec = self
                raise e

            if self.case == 'upper':
                if str_in.upper() != str_in:
                    msg = "Expected upper case string. Read '{}'.".format(
                        str_in)
                    e = InvalidEntryError(msg)
                    e.field_spec = self
                    raise e
            elif self.case == 'lower':
                if str_in.lower() != str_in:
                    msg = "Expected lower case string. Read '{}'.".format(
                        str_in)
                    e = InvalidEntryError(msg)
                    e.field_spec = self
                    raise e
            elif self.case == 'mixed':
                pass
            else:
                raise ValueError(
                    'Invalid case property {}.'.format(self.case))


class IntegerSpec(FieldSpec):
    """ Represents a field that holds integers.

        Minimum and maximum values may be specified.

        :ivar int minimum: The minimum permitted value.
        :ivar int maximum: The maximum permitted value or None.
    """

    def __init__(self,
                 identifier: str,
                 hashing_properties: FieldHashingProperties,
                 description: Optional[str] = None,
                 minimum: Optional[int] = None,
                 maximum: Optional[int] = None,
                 **kwargs: Dict[str, Any]
                 ) -> None:
        """ Make a IntegerSpec object, setting it attributes to values
            specified in keyword arguments.
        """
        # noinspection PyCompatibility,PyArgumentList
        super().__init__(identifier=identifier,
                         description=description,
                         hashing_properties=hashing_properties)

        self.minimum = minimum
        self.maximum = maximum

    @classmethod
    def from_json_dict(cls,
                       json_dict: Dict[str, Any]
                       ) -> 'IntegerSpec':
        """ Make a IntegerSpec object from a dictionary containing its
            properties.

            :param dict json_dict: This dictionary may contain
                `'minimum'` and `'maximum'` keys. In addition, it must
                contain a `'hashing'` key, whose contents are passed to
                :class:`FieldHashingProperties`.

            :param dict json_dict: The properties dictionary.
        """
        # noinspection PyCompatibility
        result = cast(IntegerSpec,  # For Mypy.
                      super().from_json_dict(json_dict))

        format_ = json_dict['format']
        result.minimum = format_.get('minimum')
        result.maximum = format_.get('maximum')

        return result

    def validate(self, str_in: Text) -> None:
        """ Validates an entry in the field.

            Raises `InvalidEntryError` iff the entry is invalid.

            An entry is invalid iff (1) the string does not represent a
            base-10 integer; (2) the integer is not between
            `self.minimum` and `self.maximum`, if those exist; or (3)
            the integer is negative.

            :param str str_in: String to validate.
            :raises InvalidEntryError: When entry is invalid.
        """
        if self.is_missing_value(str_in):
            return
        # noinspection PyCompatibility
        super().validate(str_in)

        try:
            value = int(str_in, base=10)
        except ValueError as e:
            msg = "Invalid integer. Read '{}'.".format(str_in)
            e_new = InvalidEntryError(msg)
            e_new.field_spec = self
            raise e_new from e

        if self.minimum is not None and value < self.minimum:
            msg = ("Expected integer value of at least {}. Read '{}'."
                   .format(self.minimum, value))
            e_new = InvalidEntryError(msg)
            e_new.field_spec = self
            raise e_new

        if self.maximum is not None and value > self.maximum:
            msg = ("Expected integer value of at most {}. Read '{}'."
                   .format(self.maximum, value))
            e_new = InvalidEntryError(msg)
            e_new.field_spec = self
            raise e_new

    def _format_regular_value(self, str_in: Text) -> Text:
        """ we need to reformat integer strings, as there can be different
        strings for the same integer. The
        strategy of unification here is to first parse the integer
        string to an Integer type. Thus all of
        '+13', ' 13', '13' will be parsed to 13. We then convert the
        integer value to an unambiguous string
        (no whitespaces, leading '-' for negative numbers, no leading '+').

        :param str_in: integer string
        :return: integer string without whitespaces, leading '-' for
            negative numbers, no leading '+'
        """
        try:
            value = int(str_in, base=10)
            return str(value)
        except ValueError as e:
            msg = "Invalid integer. Read '{}'.".format(str_in)
            e_new = InvalidEntryError(msg)
            e_new.field_spec = self
            raise e_new from e


class DateSpec(FieldSpec):
    """ Represents a field that holds dates.

    Dates are specified as full-dates in a format that can be described
    as a *strptime()* (C89 standard) compatible
    format string.
    E.g.: the format for the standard internet format `RFC3339
    <https://tools.ietf.org/html/rfc3339>`_
    (e.g. 1996-12-19) is '%Y-%m-%d'.

    :ivar str format: The format of the date.
    """
    OUTPUT_FORMAT = '%Y%m%d'

    def __init__(self,
                 identifier: str,
                 hashing_properties: FieldHashingProperties,
                 format: str,
                 description: Optional[str] = None
                 ) -> None:
        """ Make a DateSpec object, setting it attributes to values
            specified in keyword arguments.
        """
        # noinspection PyCompatibility,PyArgumentList
        super().__init__(identifier=identifier,
                         description=description,
                         hashing_properties=hashing_properties)

        self.format = format

    @classmethod
    def from_json_dict(cls,
                       json_dict: Dict[str, Any]
                       ) -> 'DateSpec':
        """ Make a DateSpec object from a dictionary containing its
            properties.

            :param dict json_dict: This dictionary must contain a
                `'format'` key. In addition, it must contain a
                `'hashing'` key, whose contents are passed to
                :class:`FieldHashingProperties`.

            :param json_dict: The properties dictionary.
        """
        # noinspection PyCompatibility
        result = cast(DateSpec,  # For Mypy.
                      super().from_json_dict(json_dict))

        format_ = json_dict['format']
        result.format = format_['format']

        return result

    def validate(self, str_in: Text) -> None:
        """ Validates an entry in the field.

            Raises `InvalidEntryError` iff the entry is invalid.

            An entry is invalid iff (1) the string does not represent a
            date in the correct format; or (2) the date it represents
            is invalid (such as 30 February).

            :param str str_in: String to validate.
            :raises InvalidEntryError: Iff entry is invalid.
            :raises ValueError: When self.format is unrecognised.
        """
        if self.is_missing_value(str_in):
            return
        # noinspection PyCompatibility
        super().validate(str_in)
        try:
            datetime.strptime(str_in, self.format)
        except ValueError as e:
            msg = "Validation error for date type: {}".format(e)
            e_new = InvalidEntryError(msg)
            e_new.field_spec = self
            raise e_new from e

    def _format_regular_value(self, str_in: Text) -> Text:
        """ we overwrite default behaviour as we want to hash the numbers
        only, no fillers like '-', or '/'

        :param str str_in: date string
        :return: str date string with format DateSpec.OUTPUT_FORMAT
        """
        try:
            dt = datetime.strptime(str_in, self.format)
            return datetime.strftime(dt, DateSpec.OUTPUT_FORMAT)
        except ValueError as e:
            msg = "Unable to format date value '{}'. Reason: {}".format(str_in,
                                                                        e)
            e_new = InvalidEntryError(msg)
            e_new.field_spec = self
            raise e_new from e


class EnumSpec(FieldSpec):
    """ Represents a field that holds an enum.

    The finite collection of permitted values must be specified.

    :ivar values: The set of permitted values.
    """

    def __init__(self,
                 identifier: str,
                 hashing_properties: FieldHashingProperties,
                 values: Iterable[str],
                 description: Optional[str] = None
                 ) -> None:
        """ Make a EnumSpec object, setting it attributes to values
            specified in keyword arguments.
        """
        # noinspection PyCompatibility,PyArgumentList
        super().__init__(identifier=identifier,
                         description=description,
                         hashing_properties=hashing_properties)

        self.values = set(values)

    @classmethod
    def from_json_dict(cls,
                       json_dict: Dict[str, Any]
                       ) -> 'EnumSpec':
        """ Make a EnumSpec object from a dictionary containing its
            properties.

            :param dict json_dict: This dictionary must contain an
                `'enum'` key specifying the permitted values. In
                addition, it must contain a `'hashing'` key, whose
                contents are passed to :class:`FieldHashingProperties`.
        """
        # noinspection PyCompatibility
        result = cast(EnumSpec,  # Appease the gods of Mypy.
                      super().from_json_dict(json_dict))

        format_ = json_dict['format']
        result.values = set(format_['values'])

        return result

    def validate(self, str_in: Text) -> None:
        """ Validates an entry in the field.

            Raises `InvalidEntryError` iff the entry is invalid.

            An entry is invalid iff it is not one of the permitted
            values.

            :param str str_in: String to validate.
            :raises InvalidEntryError: When entry is invalid.
        """
        if self.is_missing_value(str_in):
            return
        # noinspection PyCompatibility
        super().validate(str_in)

        if str_in not in self.values:
            msg = ("Expected enum value to be one of {}. Read '{}'."
                   .format(list(self.values), str_in))
            e = InvalidEntryError(msg)
            e.field_spec = self
            raise e


class Ignore(FieldSpec):
    """
    represent a field which will be ignored throughout the clk processing.
    """

    def __init__(self,
                 identifier: Optional[str] = None
                 ) -> None:
        # noinspection PyCompatibility
        super().__init__('' if identifier is None else identifier,
                         None)

    def validate(self, str_in: Text):
        pass


# Map type string (as defined in master schema) to
FIELD_TYPE_MAP = {
    'string': StringSpec,
    'integer': IntegerSpec,
    'date': DateSpec,
    'enum': EnumSpec,
}


def spec_from_json_dict(
        json_dict: Dict[str, Any]
) -> FieldSpec:
    """ Turns a dictionary into the appropriate FieldSpec object.

        :param dict json_dict: A dictionary with properties.
        :raises InvalidSchemaError:
        :returns: An initialised instance of the appropriate FieldSpec
            subclass.
    """
    try:
        if json_dict.get('ignored', False):
            return Ignore(json_dict['identifier'])
        type_str = json_dict['format']['type']
        spec_type = cast(FieldSpec, FIELD_TYPE_MAP[type_str])
    except KeyError as e:
        raise InvalidSchemaError("the feature definition {} is incomplete. Must contain: {}".format(json_dict, e))
    return spec_type.from_json_dict(json_dict)



from setuptools import setup, find_packages

requirements = [
        "bashplotlib>=0.6.5",
        "bitarray-hardbyte>=1.0.0",             # Fork of bitarray distributing binary wheels #153
        "cryptography>=2.3",
        "jsonschema>=3.0.2",
        "mypy_extensions>=0.3",
        "pyblake2>=1.1.1; python_version < '3.6'",
        "tqdm>=4.24",
    ]

# But on Windows, something wrong happens when creating the package
# and using codecs.open. related issues are https://github.com/di/markdown-description-example/issues/4
# But this is not exactly the same either...
with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()

setup(
    name="clkhash",
    version='0.16.0',
    description='Encoding utility to create Cryptographic Linkage Keys',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/data61/clkhash',
    license='Apache',
    install_requires=requirements,
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'pytest-cov', 'requests-mock', 'codecov', 'nbval', 'hypothesis'],
    packages=find_packages(exclude=['tests']),
    package_data={'clkhash': ['data/*.csv', 'data/*.json', 'schemas/*.json']},
    project_urls={
        'Documentation': 'http://clkhash.readthedocs.io/',
        'Source': 'https://github.com/data61/clkhash',
        'Tracker': 'https://github.com/data61/clkhash/issues',
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Security :: Cryptography",
        "Environment :: Console",
        "Operating System :: OS Independent",
        "Natural Language :: English",
    ],
)



# -*- coding: utf-8 -*-

import abc
from typing import Iterable, Text, Dict, Any, Optional


class AbstractComparison(object, metaclass=abc.ABCMeta):
    """ Abstract base class for all comparisons """

    @abc.abstractmethod
    def tokenize(self, word: Text) -> Iterable[Text]:
        """ The tokenization function.

        Takes a string and returns an iterable of tokens (as strings). This should be
        implemented in a way that the intersection of two sets of tokens produced by this function approximates
        the desired comparison criteria.

        :param word: The string to tokenize.
        :return: Iterable of tokens.
        """
        pass


class NgramComparison(AbstractComparison):
    """ Enables 'n'-gram comparison for approximate string matching. An n-gram is a contiguous sequence of n items
        from a given text.

        For Example: the 2-grams of 'clkhash' are ' c', 'cl', 'lk', 'kh', 'ha', 'as', 'sh', 'h '. Note the white-
        space in the first and last token. They serve the purpose to a) indicate the beginning and end of a word, and b)
        gives every character in the input text a representation in two tokens.

        'n'-gram comparison of strings is tolerant to spelling mistakes, e.g., the strings 'clkhash' and 'clkhush' have
        6 out of 8 2-grams in common. One wrong character will affect 'n' 'n'-grams. Thus, the larger you choose 'n',
        the more the error propagates.

        A positional n-gram also encodes the position of the n-gram within the word. The positional 2-grams of
        'clkhash' are '1  c', '2 cl', '3 lk', '4 kh', '5 ha', '6 as', '7 sh', '8 h '. Positional n-grams can be useful
        for comparing words where the position of the characters are important, e.g., postcodes or phone numbers.

        :ivar n: the n in n-gram, non-negative integer
        :ivar positional: enables positional n-gram tokenization
        """

    def __init__(self, n: int, positional: Optional[bool] = False) -> None:
        if n < 0:
            raise ValueError('`n` in `n`-gram must be non-negative.')
        self.n = n
        self.positional = positional

    def tokenize(self, word: Text) -> Iterable[Text]:
        """ Produce `n`-grams of `word`.

        :param word: The string to tokenize.
        :return: Iterable of n-gram strings.
        """
        if len(word) == 0:
            return tuple()

        if self.n > 1:
            word = '{}{}{}'.format(' ' * (self.n - 1), word, ' ' * (self.n - 1))

        if self.positional:
            # These are 1-indexed.
            return ('{} {}'.format(i + 1, word[i:i + self.n])
                    for i in range(len(word) - self.n + 1))
        else:
            return (word[i:i + self.n] for i in range(len(word) - self.n + 1))

    def __repr__(self):
        return 'NgramComparison(n={}, positional={})'.format(self.n, self.positional)


class ExactComparison(AbstractComparison):
    """ Enables exact comparisons

    High similarity score if inputs are identical, low otherwise.

    Internally, this is done by treating the whole input as one token. Thus, if you have chosen the 'bitsPerToken'
    strategy for hashing, you might want to adjust the value such that the corresponding feature gets an appropriate
    representation in the filter.
    """

    def tokenize(self, word: Text) -> Iterable[Text]:
        if len(word) == 0:
            return tuple()
        else:
            return word,


class NumericComparison(AbstractComparison):
    """ enables numerical comparisons of integers or floating point numbers.

    The numerical distance between two numbers relate to the similarity of the tokens produces by this comparison class.
    We implemented the idea of Vatsalan and Christen (Privacy-preserving matching of similar patients, Journal of
    Biomedical Informatics, 2015).

    The basic idea is to encode a number's neighbourhood such that the neighbourhoods of close numbers overlap.
    For example, the neighbourhood of x=21 is 19, 20, 21, 22, 23, and the neighbourhood of y=23 is 21, 22, 23, 24, 25.
    These two neighbourhoods share three elements. The overlap of the neighbourhoods of two numbers increases the closer
    the numbers are to each other.

    There are two parameters to control the overlap.

    - `threshold_distance`: the maximum distance which leads to an non-empty overlap. Neighbourhoods for points which
                          are further apart have no elements in common. (*)

    - `resolution`: controls how many tokens are generated. (the `b` in the paper). Given an interval of size
                    `threshold_distance` we create 'resolution tokens to either side of the mid-point plus one token for
                    the mid-point. Thus, 2 * `resolution` + 1 tokens in total. A higher resolution differentiates better
                    between different values, but should be chosen such that it plays nicely with the overall Bloom
                    filter size and insertion strategy.

    (*) the reality is a bit more tricky. We first have to quantize the inputs to multiples of `threshold_distance` /
    (2 * `resolution`), in order to get comparable neighbourhoods.
    For example, if we choose a `threshold_distance` of 8 and a `resolution` of 2, then, without quantization, the
    neighbourhood of x=25 would be [21, 23, 25, 27, 29] and for y=26 [22, 24, 26, 28, 30], resulting in no overlap.
    The quantization ensures that the inputs are mapped onto a common grid. In our example, the values would be
    quantized to even numbers (multiples of 8 / (2 * 2) = 2). Thus x=25 would be mapped to 26.
    The quantization has the side effect that sometimes two values which are further than `threshold_distance` but not
    more than `threshold_distance` + 1/2 quantization level apart can share a common token. For instance, a=24.99 would
    be mapped to 24 with a neighbourhood of [20, 22, 24, 26, 28], and b=16 neighbourhood is [12, 14, 16, 18, 20].

    We produce the output tokens based on the neighbourhood in the following way. Instead of creating a neighbourhood
    around the quantized input with values dist_interval = `threshold_distance` / (2 * `resolution`) apart, we instead
    multiply all values by (2 * `resolution`). This saves the division, which can introduce numerical inaccuracies.
    Thus, the tokens for x=25 are [88, 96, 104, 112, 120].

    We are dealing with floating point numbers by quantizing them to integers by multiplying them with
    10 ** `fractional_precision` and then rounding them to the nearest integer.

    Thus, we don't support to full range of floats, but the subset between
    2.2250738585072014e-(308 - fractional_precision - log(resolution, 10)) and
    1.7976931348623157e+(308 - fractional_precision - log(resolution, 10))

    :ivar threshold_distance: maximum detectable distance. Points that are further apart won't have tokens in common.
    :ivar resolution: controls the amount of generated tokens. Total number of tokens will be 2 * resolution + 1
    :ivar fractional_precision: number of digits after the point to be considered
    """

    def __init__(self, threshold_distance: float, resolution: int, fractional_precision: int = 0) -> None:
        # check that there is enough precision to have non-zero threshold_distance
        if not threshold_distance > 0:
            raise ValueError('threhold_distance has to be positive, but was {}'.format(threshold_distance))
        if resolution < 1:
            raise ValueError('resolution has to be greater than zero, but was {}'.format(resolution))
        if fractional_precision < 0:
            raise ValueError('fractional_precision cannot be less than zero, but was {}'.format(fractional_precision))
        # instead of dividing threshold distance as in the paper, we rather multiply the inputs by 'resolution' and then
        # use threshold_distance as distance_interval (saves a division which would need more precision)
        self.distance_interval = int(round(threshold_distance * pow(10, fractional_precision)))
        if self.distance_interval == 0:
            raise ValueError('not enough fractional precision to encode threshold_distance')
        self.resolution = resolution
        self.fractional_precision = fractional_precision

    def tokenize(self, word: Text) -> Iterable[Text]:
        if len(word) == 0:
            return tuple()
        try:
            v = int(word, base=10)  # we try int first, so we don't loose precision
            if self.fractional_precision > 0:
                v *= pow(10, self.fractional_precision)
        except ValueError:
            v_float = float(word)
            if self.fractional_precision > 0:
                v = int(round(v_float * pow(10, self.fractional_precision)))
            else:
                v = int(v_float)
        v = v * 2 * self.resolution
        residue = v % self.distance_interval

        if residue == 0:
            v = v
        elif residue < self.distance_interval / 2:
            v = v - residue
        else:
            v = v + (self.distance_interval - residue)
        return [str(v + i * self.distance_interval) for i in
                range(-self.resolution, self.resolution + 1)]


class NonComparison(AbstractComparison):
    """
    Non comparison.
    """

    def tokenize(self, word: Text) -> Iterable[Text]:
        """ Null tokenizer returns empty Iterable.

        FieldSpec Ignore has hashing_properties = None
        and get_tokenizer has to return something for this case,
        even though it's never called. An alternative would be to
        use an Optional[Callable]].

        :param word: not used
        :return: empty Iterable
        """
        return ('' for i in range(0))


def get_comparator(comp_desc: Dict[str, Any]) -> AbstractComparison:
    """ Creates the comparator as defined in the schema. A comparator provides a tokenization method suitable for
    that type of comparison.

    This function takes a dictionary, containing the schema definition. It returns a subclass of AbstractComparison.
    """

    typ = comp_desc.get('type', None)

    if typ == 'ngram':
        return NgramComparison(comp_desc.get('n', -1), comp_desc.get('positional'))
    elif typ == 'exact':
        return ExactComparison()
    elif typ == 'numeric':
        return NumericComparison(threshold_distance=comp_desc.get('thresholdDistance', -1),
                                 resolution=comp_desc.get('resolution', -1),
                                 fractional_precision=comp_desc.get('fractional_precision', 0))
    else:
        raise ValueError("unsupported comparison strategy: '{}'".format(typ))
